import abc
from warnings import warn

from mxcubecore.TaskUtils import task


class AbstractMCA(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        warn(
            "AbstractMCA is deprecated. Make an mxcubecore issue if still needed",
            DeprecationWarning,
        )
        self.mca = None
        self.calib_cf = []

    @task
    def read_raw_data(self, chmin, chmax):
        """Read the raw data from chmin to chmax

        Keyword Args:
            chmin (int): channel number, defaults to 0
            chmax (int): channel number, defaults to 4095

        Returns:
            list. Raw data.
        """
        pass

    @task
    def read_roi_data(self):
        """Read the data for the predefined ROI

        Returns:
            list. Raw data for the predefined ROI channels.
        """
        pass

    @abc.abstractmethod
    @task
    def read_data(self, chmin, chmax, calib):
        """Read the data

        Keyword Args:
            chmin (float): channel number or energy [keV]
            chmax (float): channel number or energy [keV]
            calib (bool): use calibration, defaults to False

        Returns:
            numpy.array. x - channels or energy (if calib=True), y - data.
        """
        pass

    @task
    def set_calibration(self, fname=None, calib_cf=[0, 1, 0]):
        """Set the energy calibration. Give a filename or a list of calibration factors.

        Kwargs:
            fname (str): optional filename with the calibration factors
            calib_cf (list): optional list of calibration factors

        Returns:
            list. Calibration factors.

        Raises:
            IOError, ValueError
        """
        pass

    @abc.abstractmethod
    @task
    def get_calibration(self):
        """Get the calibration factors list

        Returns:
            list. Calibration factors.
        """
        pass

    @task
    def set_roi(self, emin, emax, **kwargs):
        """Configure a ROI

        Args:
            emin (float): energy [keV] or channel number
            emax (float): energy [keV] or channel number

        Keyword Args:
            element (str): element name as in periodic table
            atomic_nb (int): element atomic number
            channel (int): output connector channel number (1-8)

        Returns:
            None

        Raises:
            KeyError
        """
        pass

    @abc.abstractmethod
    @task
    def get_roi(self, **kwargs):
        """Get ROI settings

        Keyword Args:
            channel (int): output connector channel number (1-8)

        Returns:
            dict. ROI dictionary.
        """
        pass

    @task
    def clear_roi(self, **kwargs):
        """Clear ROI settings

        Keyword Args:
           channel (int): optional output connector channel number (1-8)

        Returns:
           None
        """
        pass

    @task
    def get_times(self):
        """Return a dictionary with the preset and elapsed real time [s],
        elapsed live time (if possible) [s] and the dead time [%].

        Returns:
            dict. Times dictionary.

        Raises:
            RuntimeError
        """
        pass

    @task
    def get_presets(self, **kwargs):
        """Get the preset parameters

        Keyword Args:
            ctime (float): Real time
            erange (int): energy range
            fname (str): filename where the data are stored
            ...

        Returns:
            dict.

        Raises:
            RuntimeError
        """
        pass

    @task
    def set_presets(self, **kwargs):
        """Set presets parameters

        Keyword Args:
           ctime (float): real time [s]
           erange (int): the energy range
           fname (str): file name (full path) to save the raw data

        Returns:
           None
        """
        pass

    @abc.abstractmethod
    @task
    def start_acq(self, cnt_time=None):
        """Start new acquisition. If cnt_time is not specified, counts for preset real time.

        Keyword Args:
            cnt_time (float, optional): count time [s]; 0 means to count indefinitely.

        Returns:
            None
        """
        pass

    @abc.abstractmethod
    @task
    def stop_acq(self):
        """Stop the running acquisition"""
        pass

    @task
    def clear_spectrum(self):
        """Clear the acquired spectrum"""
        pass



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import gevent

from mxcubecore.Command.Tango import DeviceProxy
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector


class P11EigerDetector(AbstractDetector):
    def __init__(self, *args):
        AbstractDetector.__init__(self, *args)

    def init(self):
        AbstractDetector.init(self)

        self.eiger_devname = self.get_property("eiger_device")
        self.filewriter_name = self.get_property("filewriter_device")
        self._roi_mode = self.get_property("roi_mode", "disabled")

        # self.log.debug("EIGER - device name is : %s" % self.eiger_devname)
        # self.log.debug("EIGER - filewriter name is : %s" % self.filewriter_name)
        # self.log.debug("EIGER - detector distance is : %s" % self._distance_motor_hwobj)
        # self.log.debug("EIGER - shutterless : %s" % self.has_shutterless())

        self.eiger_dev = DeviceProxy(self.eiger_devname)
        self.writer_dev = DeviceProxy(self.filewriter_name)

        # Get the detector distance device
        self.detector_tower_dev = DeviceProxy(
            self.get_property("detector_tower_device")
        )
        self.dcm_energy_dev = DeviceProxy(self.get_property("dcm_energy_device"))

        self.chan_status = self.get_channel_object("_status")
        self.chan_status.connect_signal("update", self.status_changed)

        self._exposure_time_limits = eval(self.get_property("exposure_time_limits"))

        self._beam_centre = eval(self.get_property("beam_centre"))
        self.originx = self._beam_centre[0]
        self.originy = self._beam_centre[1]

        self.log.debug("EIGER - beamcenter: %s" % str(self._beam_centre))
        self.log.debug("EIGER - originx: %s" % str(self.originx))
        self.log.debug("EIGER - originy: %s" % str(self.originy))

        self.inited = False

    def init_acquisition(self):
        #  set parameters like exts, images_per_file, etc...
        # self.roi_mode = "4M"
        # self.roi_mode = "disabled"

        self.set_eiger_enum("TriggerMode", "exts")
        self.set_eiger_enum("RoiMode", self._roi_mode)
        self.set_eiger_enum("Compression", "bslz4")
        self.set_writer_enum("Mode", "enabled")
        # self.writer_dev.write_attribute("NImagesPerFile", 1000)
        self.inited = True

    def has_shutterless(self):
        return True

    def prepare_acquisition(self, *args, **kwargs):
        """
        Prepares detector for acquisition
        """
        self.log.debug("EIGER - prepare_acquisition")
        # Sets the metadata for the header
        self.set_metadata()

    def get_radius(self, distance=None):
        # a proper calculation should be done here
        # this value comes from crystalControlMaxwell hardcoded value to
        # estimate resolution
        return 311 / 2.0

    def get_eiger_detector_distance(self):
        # Get current detector distance from detector tower server
        det_tower_distance = self.detector_tower_dev.read_attribute(
            "DetectorDistance"
        ).value
        self.log.debug("EIGER: DetectorTower distance is : %s" % det_tower_distance)
        return det_tower_distance

    def set_eiger_detector_distance(self):
        # Set detector distance in the Detector server for the header.
        # It does not set the actual detector distance!

        # in mm from detector tower server
        current_det_tower_distance = self.get_eiger_detector_distance()
        # in meters for the detector header info
        self.eiger_dev.write_attribute(
            "DetectorDistance", float(current_det_tower_distance / 1000.0)
        )  # in m
        # self.eiger_dev.write_attribute("DetectorDistance", float(current_det_tower_distance)) #in mkm

    # TODO add if needed
    def get_eiger_beam_center(self):
        return True

    def set_eiger_beam_center(self):
        # ==== Setting the beam center ======================
        # Beam center is depending on the detector distance as for the CrystalControl.
        # Emulate of the same hardcoded config parameters as before.
        # originx and originy (set in eiger.xml) are the X and Y of the beam mark at the detector distance of 160 mm.

        # Beam center params after calibration in October 2022 from CC:
        current_det_tower_distance = self.get_eiger_detector_distance()
        corrected_originx = (
            self.originx + 63.5 * (current_det_tower_distance - 160) / 1000.0
        )
        corrected_originy = (
            self.originy - 14.2 * (current_det_tower_distance - 160) / 1000.0
        )

        self.eiger_dev.write_attribute("BeamCenterX", float(corrected_originx))
        self.eiger_dev.write_attribute("BeamCenterY", float(corrected_originy))

        self.log.debug(
            "EIGER - current beamcenter X and Y: %f, %f at detector distance %f mm"
            % (corrected_originx, corrected_originy, current_det_tower_distance)
        )

    # TODO add if needed
    def get_eiger_photon_energy(self):
        return True

    # FIXME: Add the same conditions as for CC
    def set_eiger_photon_energy(self):
        # Sets photon energy in the detector server for the header
        current_dcm_energy = float(self.dcm_energy_dev.read_attribute("Position").value)

        if current_dcm_energy > 5500:
            self.eiger_dev.write_attribute("PhotonEnergy", float(current_dcm_energy))
        else:
            self.eiger_dev.write_attribute("PhotonEnergy", float(5500))

    def set_eiger_start_angle(self, arg):
        # Sets Detector start angle for the header
        arg = float(arg)
        self.eiger_dev.write_attribute("OmegaStart", arg)

    def set_eiger_angle_increment(self, arg):
        # Sets detector angle increment
        arg = float(arg)
        self.eiger_dev.write_attribute("OmegaIncrement", arg)

    def prepare_common(self, exptime, filepath):
        if not self.inited:
            self.init_acquisition()

        self.eiger_dev.write_attribute("CountTime", float(exptime))
        self.eiger_dev.write_attribute("FrameTime", float(exptime))
        self.eiger_dev.write_attribute("TriggerStartDelay", 0.003)

        if filepath.startswith("/gpfs"):
            filepath = filepath[len("/gpfs") :]

        self.writer_dev.write_attribute("NamePattern", filepath)

        # Sets the metadata for the header
        self.set_metadata()

    def prepare_characterisation(self, exptime, number_of_images, angle_inc, filepath):
        self.writer_dev.write_attribute(
            "NImagesPerFile", 1
        )  # To write one image per characterisation.
        self.prepare_common(exptime, filepath)
        self.eiger_dev.write_attribute("Nimages", 1)  # Number of images per trigger
        self.log.debug(
            "Eiger. preparing characterization. Number of triggers is: %d"
            % number_of_images
        )
        self.eiger_dev.write_attribute("Ntrigger", int(number_of_images))

    def prepare_std_collection(self, exptime, number_of_images, filepath):
        self.prepare_common(exptime, filepath)
        self.eiger_dev.write_attribute("Nimages", int(number_of_images))
        self.eiger_dev.write_attribute("Ntrigger", 1)
        self.writer_dev.write_attribute("NImagesPerFile", 1000)  # Default

        self.writer_dev.write_attribute("ImageNrStart", 1)

    def set_metadata(self):
        self.set_eiger_detector_distance()  # set detector distance for the header
        self.set_eiger_beam_center()  # set detector beam center for the header
        self.set_eiger_photon_energy()  # set detector photon energy

    def start_acquisition(self):
        self.eiger_dev.Arm()
        self.wait_ready()

    def stop_acquisition(self):
        self.eiger_dev.Abort()
        gevent.sleep(1)
        self.eiger_dev.Disarm()
        self.wait_ready()

    def wait_ready(self, timeout=30):
        with gevent.Timeout(timeout, RuntimeError("timeout waiting detector ready")):
            while self.chan_status.get_value().lower() not in ["ready", "idle"]:
                gevent.sleep(2)

    def status_changed(self, status):
        self.log.debug("P11EigerDetector - status changed. now is %s" % status)

    def get_beam_position(self, distance=None, wavelength=None):
        return self._beam_centre

    # managing enum device server attributes
    def set_eiger_enum(self, attr, value):
        dev = self.eiger_dev
        self.set_attr_enum(dev, attr, value)

    def set_writer_enum(self, attr, value):
        dev = self.writer_dev
        self.set_attr_enum(dev, attr, value)

    def set_attr_enum(self, dev, attr, value):
        values = list(dev.get_attribute_config(attr).enum_labels)

        no = values.index(value)
        if no >= 0:
            print("writing no %s for value %s (%s)" % (no, value, str(values)))
            dev.write_attribute(attr, no)
        else:
            self.log.error(
                "Trying to write invalid value %s for attribute %s" % (value, attr)
            )

    def get_eiger_enum(self, attr):
        dev = self.eiger_dev
        return self.get_attr_enum(dev, attr)

    def get_writer_enum(self, attr):
        dev = self.writer_dev
        return self.get_attr_enum(dev, attr)

    def get_attr_enum(self, dev, attr):
        values = list(dev.get_attribute_config(attr).enum_labels)
        no = dev.read_attribute(attr).value
        return values[no]

    def get_eiger_name_pattern(self):
        return self.writer_dev.NamePattern

    def get_latest_local_master_image_name(self):
        latest_image = self.get_eiger_name_pattern()
        latest_image = f"/gpfs{latest_image}_master.h5"
        return latest_image



from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ExpMotor import ExpMotor


class MicrodiffFocusMotor(ExpMotor):
    def __init__(self, name):
        ExpMotor.__init__(self, name)

    def init(self):
        if HWR.beamline.diffractometer.in_plate_mode:
            self.actuator_name = self.get_property("centring_focus")
        else:
            self.actuator_name = self.get_property("alignment_focus")
        ExpMotor.init(self)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Micro Diffractometer implementation of the AbstractDiffractometer class."""

import time

from gevent import Timeout, sleep

from mxcubecore import HardwareRepository as HWR
from mxcubecore.Command.Exporter import Exporter
from mxcubecore.Command.exporter.ExporterStates import ExporterStates
from mxcubecore.HardwareObjects import sample_centring
from mxcubecore.HardwareObjects.abstract.AbstractDiffractometer import (
    AbstractDiffractometer,
    DiffractometerConstraint,
    DiffractometerHead,
    DiffractometerPhase,
)

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicroDiffractometer(AbstractDiffractometer):
    """Microdiff with Exporter implementation of AbstractDiffractometer"""

    def __init__(self, name):
        super().__init__(name)
        self._exporter = None

    def init(self):
        """Initialise the device"""
        super().init()
        exporter_address = self.get_property("exporter_address")
        _host, _port = exporter_address.split(":")
        self._exporter = Exporter(_host, int(_port))
        self.head_type = self._head_type

        # add the custom commands
        for nam, cmd in self.get_property("commands").items():
            _cmd = {
                "type": "exporter",
                "exporter_address": exporter_address,
                "name": nam,
            }
            setattr(self, nam, self.add_command(_cmd, cmd))

        # add the custom channels
        for nam, attr in self.get_property("channels").items():
            _attr = {
                "type": "exporter",
                "exporter_address": exporter_address,
                "name": nam,
            }
            setattr(self, nam, self.add_channel(_attr, attr))

        self.update_state()

        # we must have global_state and phase_channel channels configured
        try:
            self.phase_channel.connect_signal("update", self.update_phase)
            self.global_state.connect_signal("update", self._update_state)
        except AttributeError:
            self.log.exception("global_state and phase_channel not configured!")

        sample_centring.NUM_CENTRING_ROUNDS = self.get_property(
            "num_centering_rounds", 1
        )

    def abort(self):
        """Immediately terminate action."""
        self._exporter.execute("abort")

    @property
    def _get_hwstate(self) -> str:
        """Get the hardware state, reported by the MD2 application.

        Returns:
            The state.
        """
        try:
            return self._exporter.read_property("HardwareState")
        except AttributeError:
            return "Ready"

    @property
    def _get_swstate(self) -> str:
        """Get the software state, reported by the MD2 application.

        Returns:
            The state.
        """
        return self.global_state.get_value()

    def _update_state(self, value):
        if isinstance(value, str):
            self.update_state(self.get_state())
        else:
            self.update_state(value)

    def get_state(self):
        """Get the diffractometer general state.

        Returns:
            (enum 'HardwareObjectState'): state
        """
        try:
            return ExporterStates[self._get_swstate.upper()].value
        except ValueError:
            return self.STATES.UNKNOWN

    @property
    def _ready(self) -> bool:
        """Get the "Ready" state - software and hardware.

        Returns:
            True if both "Ready", False otherwise.
        """
        return self._get_swstate == "Ready" and self._get_hwstate == "Ready"

    def wait_status_ready(self, timeout: float | None = None):
        """Wait timeout seconds until status is ready.

        Args:
            Timeout [s]. None means infinite timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for status ready")):
            while not self._ready:
                sleep(0.5)

    def set_value_motors(
        self,
        motors_positions_dict: dict[str, float],
        simultaneous: bool = True,
        timeout: float | None = None,
    ):
        """Move specified motors to the requested positions.

        Args:
            motors_positions_dict (dict): Dictionary {motor_role: target_value}.
            simultaneous: Move the motors simultaneously (True - default) or not.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            TimeoutError: Timeout
            KeyError: The name does not correspond to an existing motor
        """
        if not simultaneous:
            super().set_value_motors(motors_positions_dict, simultaneous, timeout)
        else:
            # prepare the command
            cmd = ""
            for role, pos in motors_positions_dict.items():
                name = self.motors_hwobj_dict[role].actuator_name
                cmd += f"{name}={pos:0.3f};"
            self._exporter.execute("startSimultaneousMoveMotors", (cmd,))
            if timeout != 0:
                self.wait_status_ready(timeout)
            self.update_state()

    def get_value_motors(self, motors_list: [list | None] = None) -> dict[str, float]:
        """Get the positions of diffractometer motors. If the motors_list
           is empty, return the positions of all the availble motors.

        Args:
            motors_list: List of motor roles.

        Returns:
            Dictionary {role: position}.
        """
        motors_positions_dict = super().get_value_motors(motors_list)
        if not self.in_kappa_mode:
            motors_positions_dict.update({"kappa": None, "kappa_phi": None})
        return motors_positions_dict

    def get_motors(self):
        """Get the dictionary of all the motors which can be used.

        Returns:
            Ddctionary {role: hardware_object}.
        """

        def find_elem(ddict, val):
            """Find dictionary elemnt from motor actuator_name"""
            for role, hwobj in ddict.items():
                if hwobj.actuator_name == val:
                    return {role: hwobj}
            return {}

        motors = self._exporter.read_property("MotorStates")
        motors_dict = {}
        for mot in motors:
            mot_stat = mot.split("=")
            if mot_stat[1] not in ("Disable", "Unknown"):
                elem = find_elem(self.motors_hwobj_dict, mot_stat[0])
                motors_dict.update(elem)

        return motors_dict

    @property
    def _head_type(self) -> DiffractometerHead:
        """Get the head type."""
        try:
            self.head_type = DiffractometerHead(
                self._exporter.read_property("HeadType")
            )
        except ValueError:
            self.head_type = DiffractometerHead.UNKNOWN
        return self.head_type

    def _set_phase(self, value: DiffractometerPhase):
        """Specific implementation to set the diffractometer to selected phase.

        Args:
            value: requested phase.
        """
        _use_custom = self.get_property("use_custom_phase_script") or False
        current_phase = self.get_phase()

        if value != current_phase:
            msg = f"Current phase is {current_phase} and moving to {value}"
            self.log.info(msg)

        # protect the detector or open the cover if detector cover defined
        # do not wait it to finish.
        det_cover = self.get_object_by_role("detector_cover")
        if det_cover:
            if value in (DiffractometerPhase.TRANSFER, DiffractometerPhase.SEE_BEAM):
                det_cover.set_value(det_cover.detector_cover.VALUES.CLOSE, timeout=0)
            if value == DiffractometerPhase.COLLECT:
                det_cover.set_value(det_cover.detector_cover.VALUES.OPEN, timeout=0)

        if _use_custom and not self.in_plate_mode:
            script = "ChangePhase_" + value.value.lower()
            msg = f"Changing phase to {value.value}, using pmac script"
            self.log.info(msg)
            self.run_custom_script(script)
        else:
            self._exporter.execute("startSetPhase", (value.value,))
        self.wait_status_ready(timeout=600)

    def get_phase(self) -> DiffractometerPhase:
        """Get the current phase."""
        value = self.phase_channel.get_value()
        try:
            self.current_phase = DiffractometerPhase(value)
        except ValueError:
            self.current_phase = DiffractometerPhase.UNKNOWN
        return self.current_phase

    def _set_constraint(self, value: DiffractometerConstraint):
        """Specific implementation to set the diffractometer to selected constraint
        Args:
            value: requested constraint.
        """
        self._exporter.execute("startSetMode", (value.value,))

    def get_constraint(self) -> DiffractometerConstraint:
        """Get the diffrractometer constraint type."""
        value = self._exporter.read_property("CurrentMode")
        try:
            self.current_constraint = DiffractometerConstraint(value)
        except ValueError:
            self.current_constraint = DiffractometerConstraint.UNKNOWN
        return self.current_constraint

    def check_scan_limits(self, start: float, end: float, exptime: float) -> bool:
        """Check if the scan parameters are within the limits

        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total) [s].

        Returns:
            True (parameters within the limits), False otherwise.
        """
        if self.in_plate_mode:
            scan_speed = abs(end - start) / exptime
            llim, hlim = map(
                float,
                self._exporter.execute("getOmegaMotorDynamicScanLimits", (scan_speed,)),
            )
            if start < llim:
                msg = f"Scan start below the allowed value {llim}"
                raise ValueError(msg)
            if end > hlim:
                msg = f"Scan end above the allowed value {hlim}"
                raise ValueError(msg)
        return True

    def do_oscillation_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        number_of_images: int = 1,
        timeout: float | None = None,
    ):
        """Do an oscillation scan on omega.

        Args:
            start: omega start position.
            end: omega end position.
            exptime: scan exposure time (total).
            number_of_images: Used if need to set number of frames.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
            ValueError: Scan parameters not within limits (if relevant).
        """
        # check the scan limits
        self.check_scan_limits(start, end, exptime)
        # set the number of frames
        if not self.get_property("md_set_number_of_frames"):
            number_of_images = 1
        self._exporter.write_property("ScanNumberOfFrames", number_of_images)
        scan_params = f"1\t{start:0.3f}\t{(end - start):0.3f}\t{exptime:0.3f}\t1"
        self._exporter.execute("startScanEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_line_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        number_of_images: int,
        motors_pos: dict[str, dict],
        timeout: float | None = None,
    ):
        """Do helical (line) scan on omega.
        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total).
            number_of_images: Used only if more tahn one frame needed.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).
            motors_pos: {"1": [centred position], "2": [centred position]}

        Raises:
            RuntimeError: Timeout waiting for status ready.
            ValueError: Scan parameters not within limits (if relevant).
        """
        # check the scan limits
        self.check_scan_limits(start, end, exptime)
        if not self.get_property("md_set_number_of_frames"):
            number_of_images = 1

        self._exporter.write_property("ScanNumberOfFrames", number_of_images)

        scan_params = f"{start:0.3f}\t{(end - start):0.3f}\t{exptime:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            scan_params += f"{motors_pos['1'][name]:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            scan_params += f"{motors_pos['2'][name]:0.3f}\t"

        self._exporter.execute("startScan4DEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_mesh_scan(
        self,
        start: float,
        end: float,
        exptime: float,
        dead_time: float,
        nb_lines: int,
        nb_frames_total: int,
        grid_centre: list[tuple[str, float]],
        mesh_range: dict,
        timeout: float | None = None,
    ):
        """Do a mesh scan.

        Args:
            start: scan start position.
            end: scan end position.
            exptime: scan exposure time (total).
            dead_time: Dead time between the pulses. Detector dependant.
            nb_lines: Total number of lines.
            nb_frames_total: Total number of frames.
            grid_centre: List of tuples (motor_role, position).
                         representing the centre of the mesh grid.
            mesh_range: Horizontal and vertical range.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """

        # enable gate pulses
        self._exporter.write_property("DetectorGatePulseEnabled", value=True)

        # dead_time depends on the detector. We transform it to us
        dead_time = dead_time or HWR.beamline.detector.get_deadtime() * 1000
        self._exporter.write_property("DetectorGatePulseReadoutTime", dead_time)

        grid_centre = grid_centre.as_dict()
        self.set_value_motors(grid_centre, simultaneous=True, timeout=timeout)

        scan_params = f"{(end - start):0.3f}\t"
        scan_params += f"{-mesh_range['horizontal_range']:0.3f}\t"
        scan_params += f"{mesh_range['vertical_range']:0.3f}\t"
        scan_params += f"{start:0.3f}\t"
        for name in ["phiy", "phiz", "sampx", "sampy"]:
            for key, val in grid_centre.items():
                if name == key:
                    scan_params += f"{float(val):0.3f}\t"
        scan_params += f"{nb_lines}\t"
        scan_params += f"{nb_frames_total / nb_lines}\t"
        scan_params += f"{exptime / nb_lines}\t"
        scan_params += "True\tTrue\tTrue\t"
        self._exporter.execute("startRasterScanEx", (scan_params,))
        self.wait_status_ready(timeout)

    def do_still_scan(
        self,
        pulse_duration: float,
        pulse_period: float,
        nb_pulse: int,
        timeout: [None | float] = None,
    ):
        """Do a zero oscillation acquisition.

        Args:
            pulse_duration: Duration of the pulse sent to the detector.
            pulse_period: The period of the pulse sent to the detector.
            nb_pulse: Number of pulses to be sent.
            timeout: optional - timeout [s],
                     if timeout = 0: return at once and do not wait,
                     if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """
        scan_params = f"{pulse_duration:0.6f}\t{pulse_period:0.6f}\t{nb_pulse}"
        self._exporter.execute("startStillScan", (scan_params,))
        self.wait_status_ready(timeout)

    def do_characterisation_scan(
        self,
        start: float,
        scan_range: float,
        nb_frames: int,
        exptime: float,
        nb_scans: int,
        angle: float,
        timeout: float | None = None,
    ):
        """Do fast characterisation.
        Args:
            start: Position of omega for the first scan [deg].
            scan_range: range for each scan [deg].
            nb_frames: Frame numbers for each scan.
            exptime: Total exposure time for each scan [s].
            nb_scans: How many times a scan to be repeated.
            angle: The angle between each scan [deg]. This number,
                   added to the last position of each scan and will
                   be the start position of the following scan.
            timeout (float): optional - timeout [s],
                             if timeout = 0: return at once and do not wait,
                             if timeout is None: wait forever (default).

        Raises:
            RuntimeError: Timeout waiting for status ready.
        """

        if self.in_plate_mode:
            # to see if needed when plates
            return

        scan_params = f"{nb_frames}\t{start:0.3f}\t{scan_range:0.3f}\t"
        scan_params += f"{exptime:0.3f}\t{nb_scans}\t{angle:0.3f}"
        self._exporter.execute("startCharacterisationScanEx", (scan_params,))
        if timeout:
            # min timeout is 20 min
            timeout = max(timeout, 20 * 60)

        self.wait_status_ready(timeout)

    def get_pixels_per_mm(self) -> tuple[int, int]:
        """Get the pixel/mm values.

        Returns:
            (x ,y) [pixel/mm]
        """
        x_calib = self._exporter.read_property("CoaxCamScaleX")
        y_calib = self._exporter.read_property("CoaxCamScaleY")
        return 1.0 / x_calib, 1.0 / y_calib

    def get_beam_position(self) -> tuple:
        """Get the beam position defined in MD"""
        return (
            self.beam_position_horizontal.get_value(),
            self.beam_position_vertical.get_value(),
        )

    def run_custom_script(self, script_cmd: str, timeout: float | None = None):
        """Run custom script."""
        self.run_script(script_cmd)
        # Wait for script to start before checking status,
        # can perhaps be improved to wait for ready-busy-ready ?
        time.sleep(0.4)
        self.wait_status_ready(timeout)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""FlexHCD Linux Java implementation of sample changer.
Example xml file:
<object class = "EMBLFlexHCD">
  <username>Sample Changer</username>
  <exporter_address>lid231flex1:9001</exporter_address>
  <object role="controller" href="/bliss"/>
  <puck_configuration>["SC3", "UNI", "SC3", "UNI", "UNI", "UNI", "UNI", "UNI"]<
/puck_configuration>
</object>
"""

import ast
import base64
import logging
import pickle
import time
from typing import (
    Any,
    List,
)

import gevent
from PyTango.gevent import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    SampleChanger,
    SampleChangerState,
)
from mxcubecore.HardwareObjects.abstract.sample_changer.Container import (
    Container,
    Sample,
)
from mxcubecore.TaskUtils import task


class Pin(Sample):
    def __init__(self, basket, cell_no, basket_no, sample_no):
        super(Pin, self).__init__(
            basket, Pin.get_sample_address(cell_no, basket_no, sample_no), True
        )
        self._set_holder_length(22.0)
        self.present = True

    def get_basket_no(self):
        return self.get_container().get_index() + 1

    def get_vial_no(self):
        return self.get_index() + 1

    def get_cell_no(self):
        return self.get_container().get_container().get_index() + 1

    def get_cell(self):
        return self.get_container().get_container()

    @staticmethod
    def get_sample_address(cell_number, basket_number, sample_number):
        return (
            str(cell_number) + ":" + str(basket_number) + ":" + "%02d" % sample_number
        )


class Basket(Container):
    __TYPE__ = "Puck"

    def __init__(self, container, cell_no, basket_no, unipuck=False):
        super(Basket, self).__init__(
            self.__TYPE__,
            container,
            Basket.get_basket_address(cell_no, basket_no),
            True,
        )
        for i in range(16 if unipuck else 10):
            slot = Pin(self, cell_no, basket_no, i + 1)
            self._add_component(slot)
        self.present = True

    @staticmethod
    def get_basket_address(cell_number, basket_number):
        return str(cell_number) + ":" + str(basket_number)

    def get_cell_no(self):
        return self.get_container().get_index() + 1

    def get_cell(self):
        return self.get_container()

    def clear_info(self):
        self.get_container()._reset_basket_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()


class Cell(Container):
    __TYPE__ = "Cell"

    def __init__(self, container, number, puck_type="SC3"):
        super(Cell, self).__init__(
            self.__TYPE__, container, Cell.get_cell_address(number), True
        )
        self.present = True

        if puck_type == "SC3":
            for i in range(3):
                self._add_component(Basket(self, number, i + 1, unipuck=False))
        else:
            for i in range(3):
                self._add_component(Basket(self, number, i + 1, unipuck=True))

    @staticmethod
    def get_cell_address(cell_number):
        return str(cell_number)

    def _reset_basket_info(self, basket_no):
        pass

    def clear_info(self):
        self.get_container()._reset_cell_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()

    def get_cell(self):
        return self


class EMBLFlexHCD(SampleChanger):
    __TYPE__ = "Flex Sample Changer"

    def __init__(self, *args, **kwargs):
        super(EMBLFlexHCD, self).__init__(self.__TYPE__, True, *args, **kwargs)

    def init(self):
        _pucks = '["UNI", "UNI", "UNI", "UNI", "UNI", "UNI", "UNI", "UNI"]'
        pucks = ast.literal_eval(self.get_property("puck_configuration", _pucks))

        for i in range(8):
            cell = Cell(self, i + 1, pucks[i])
            self._add_component(cell)

        self.robot = self.get_property("tango_device")
        if self.robot:
            self.robot = DeviceProxy(self.robot)

        self.exporter_addr = self.get_property("exporter_address")

        self.swstate_attr = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "swstate",
            },
            "State",
        )

        self.controller = self.get_object_by_role("controller")
        self.prepareLoad = self.get_command_object("moveToLoadingPosition")
        self.timeout = 3
        self.gripper_types = {
            -1: "No Gripper",
            1: "UNIPUCK",
            2: "MINISPINE",
            3: "FLIPPING",
            4: "UNIPUCK_DOUBLE",
            5: "PLATE",
        }

        SampleChanger.init(self)
        # self._set_state(SampleChangerState.Disabled)
        self._update_selection()
        self.state = self._read_state()

    def get_sample_list(self) -> List[Any]:
        """
        Returns a list of present samples enriched with information
        from the Flex getPresentSamples() call.
        """
        sample_list = super().get_sample_list()
        present_sample_list = []
        # Get serialized Present samples
        sc_present_sample_list = self._execute_cmd_exporter(
            "getPresentSamples", attribute=True
        )

        if not sc_present_sample_list:
            return []

        # Parse colon-separated entries
        entries = [e.strip() for e in sc_present_sample_list.split(":") if e.strip()]

        # Convert entries to structured data
        # and match parsed Flex samples with MXCuBECore sample objects
        for sample in sample_list:
            for entry in entries:
                try:
                    cell, puck, puck_type, puck_barcode, well, sample_barcode, state = [
                        p.strip() for p in entry.split(",")
                    ]
                    state = "in_puck" if state == "on_gonio" else state
                except ValueError:
                    self.log.warning("Skipping unexpected sample entry: %s", entry)
                    continue

                sample_addr = f"{cell}:{puck}:{int(well):02d}"
                if sample.get_address() == sample_addr:
                    if sample.state != state:
                        # Add extra info directly into the sample object
                        sample.state = state
                        sample.puck_barcode = puck_barcode
                        sample.sample_barcode = sample_barcode
                        sample.puck_type = puck_type

                        self._trigger_sample_info_changed_event(sample)

                    present_sample_list.append(sample)
                    break  # stop inner loop once matched

        return present_sample_list

    @task
    def prepare_load(self):
        if self.controller:
            self.controller.hutch_actions(enter=True, hutch_trigger=True)
        else:
            self.prepareLoad()

    @task
    def _prepare_centring_task(self):
        if self.controller:
            self.controller.hutch_actions(enter=False, sc_loading=True)
        else:
            gevent.sleep(2)
            self.get_command_object("unlockMinidiffMotors")(wait=True)
            self.get_command_object("prepareCentring")(wait=True)

    def prepare_centring(self):
        self._prepare_centring_task()

    def get_sample_properties(self):
        return (Pin.__HOLDER_LENGTH_PROPERTY__,)

    def get_basket_list(self):
        basket_list = []
        # put here only the baskets that exist, not all the possible ones
        for cell in self.get_components():
            for basket in cell.get_components():
                if isinstance(basket, Basket):
                    basket_list.append(basket)

        return basket_list

    def _do_change_mode(self, *args, **kwargs):
        return

    def _do_update_info(self):
        self._update_selection()
        self._update_state()

    def _do_scan(self, component, recursive=True, saved={"barcodes": None}):
        return

    def _execute_cmd(self, cmd, *args, **kwargs):
        timeout = kwargs.pop("timeout", None)
        if args:
            cmd_str = "flex.%s(%s)" % (cmd, ",".join(map(repr, args)))
        else:
            cmd_str = "flex.%s()" % cmd
        cmd_id = self.robot.eval(cmd_str)

        if not cmd_id:
            cmd_id = self.robot.eval(cmd_str)
        with gevent.Timeout(
            timeout, RuntimeError("Timeout while executing %s" % repr(cmd_str))
        ):
            while True:
                if self.robot.is_finished(cmd_id):
                    break
                gevent.sleep(0.2)

        res = self.robot.get_result(cmd_id)
        if res:
            res = pickle.loads(base64.b64decode(res))
            if isinstance(res, Exception):
                raise res
            return res

    def _execute_cmd_exporter(self, cmd, *args, **kwargs):
        ret = None
        timeout = kwargs.pop("timeout", 900)
        if args:
            args_str = "%s" % "\t".join(map(repr, args))
        if kwargs.pop("command", None):
            exp_cmd = self.add_command(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd,
            )
            if args:
                ret = exp_cmd(args_str)
            else:
                ret = exp_cmd()
        if kwargs.pop("attribute", None):
            exp_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd[3:],
            )
            if cmd.startswith("get"):
                return exp_attr.get_value()
            if cmd.startswith("set"):
                ret = exp_attr.set_value(args_str)

        self._wait_ready(timeout=timeout)
        return ret

    def _assert_ready(self):
        if not self._ready():
            raise RuntimeError("Sample changer is busy cant mount/unmount")

    def _ready(self):
        return self._execute_cmd_exporter("getState", attribute=True) == "Ready"

    def _busy(self):
        return self._execute_cmd_exporter("getState", attribute=True) != "Ready"

    def _wait_ready(self, timeout=None):
        # None means wait forever timeout <=0 use default timeout
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for sample changer to be ready"

        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._ready():
                gevent.sleep(0.5)

    def _wait_busy(self, timeout=None):
        # None means wait forever timeout <=0 use default timeout
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for sample changer action to start"

        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._busy():
                gevent.sleep(0.5)

    def _do_select(self, component):
        if isinstance(component, Cell):
            cell_pos = component.get_index() + 1
        elif isinstance(component, (Basket, Pin)):
            cell_pos = component.get_cell_no()

        self._execute_cmd_exporter("moveDewar", cell_pos, command=True)

        self._update_selection()

    def chained_load(self, old_sample, sample):
        return self._do_load(sample)

    def _set_loaded_sample_and_prepare(self, sample, previous_sample):
        res = False

        if -1 not in sample and sample != previous_sample:
            self._set_loaded_sample(self.get_sample_with_address(sample))
            self._prepare_centring_task()
            res = True

        return res

    def _hw_get_mounted_sample(self):
        loaded_sample = tuple(
            self._execute_cmd_exporter("getMountedSamplePosition", attribute=True)
        )

        return (
            str(loaded_sample[0])
            + ":"
            + str(loaded_sample[1])
            + ":"
            + "%02d" % loaded_sample[2]
        )

    def get_sample_with_address(self, address):
        sample = None
        address = str(address[0]) + ":" + str(address[1]) + ":" + "%02d" % address[2]

        for s in self.get_sample_list():
            if s.get_address() == address:
                sample = s

        return sample

    def reset_loaded_sample(self):
        self._execute_cmd_exporter("resetLoadedPosition", command=True)
        self._reset_loaded_sample()

    def get_robot_exceptions(self):
        return [self._execute_cmd_exporter("getLastTaskException", attribute=True)] or [
            ""
        ]

    @task
    def load(self, sample):
        self.prepare_load()
        self.enable_power()

        try:
            res = SampleChanger.load(self, sample)
        finally:
            for msg in self.get_robot_exceptions():
                if msg is not None:
                    self.log.error(msg)

        # if res:
        #    self.prepare_centring()

        return res

    @task
    def unload_sample(
        self,
        holderLength,
        sample_id=None,
        sample_location=None,
        successCallback=None,
        failureCallback=None,
    ):
        self._assert_ready()
        cell, basket, sample = sample_location
        sample = self.get_component_by_address(
            Pin.get_sample_address(cell, basket, sample)
        )
        return self.unload(sample)

    @task
    def unload(self, sample):
        self.prepare_load()
        self.enable_power()

        if not sample:
            sample = self._hw_get_mounted_sample()

        try:
            SampleChanger.unload(self, sample)
        finally:
            for msg in self.get_robot_exceptions():
                if msg is not None:
                    self.log.error(msg)

    def get_gripper(self):
        gripper_type = self._execute_cmd_exporter("get_gripper_type", attribute=True)

        return self.gripper_types.get(gripper_type, "?")

    def get_available_grippers(self):
        grippers = []
        try:
            ret = sorted(
                self._execute_cmd_exporter("getSupportedGrippers", attribute=True)
            )
            for gripper in ret:
                grippers.append(self.gripper_types[gripper])
        except Exception:
            grippers = [-1]

        return grippers

    @task
    def change_gripper(self, gripper=None):
        self.prepare_load()
        self.enable_power()

        if gripper:
            self._execute_cmd_exporter("setGripper", gripper, command=True)
        else:
            self._execute_cmd_exporter("changeGripper", command=True)

    @task
    def home(self):
        self.prepare_load()
        self.enable_power()
        self._execute_cmd_exporter("homeClear", command=True)

    @task
    def enable_power(self):
        if not self.exporter_addr:
            self._execute_cmd("enablePower", 1)

    @task
    def defreeze(self):
        self.prepare_load()
        self.enable_power()
        self._execute_cmd_exporter("defreezeGripper", command=True)

    def _do_load(self, sample=None):
        self._update_state()
        previous_sample = tuple(
            self._execute_cmd_exporter("getMountedSamplePosition", attribute=True)
        )
        loaded_sample = previous_sample

        # We wait for the sample changer if it is already doing something,
        # like defreezing.
        # wait for 10 minutes then timeout !
        state = self._execute_cmd_exporter("getStatus", attribute=True)
        if state == "Defreezing Gripper":
            msg = f"Sample changer in operation ({state}), please wait"
            logging.getLogger("user_level_log").warning(msg)
        self._wait_ready(600)

        # Start loading
        load_task = gevent.spawn(
            self._execute_cmd_exporter,
            "loadSample",
            sample.get_cell_no(),
            sample.get_basket_no(),
            sample.get_vial_no(),
            command=True,
        )

        # Wait for sample changer to start activity
        try:
            _tt = time.time()
            self._wait_busy(300)
            self.log.info(f"Waited SC activity {time.time() - _tt}")
        except Exception:
            for msg in self.get_robot_exceptions():
                logging.getLogger("user_level_log").error(msg)
            raise

        # Wait for the sample to be loaded, (put on the goniometer)
        err_msg = "Timeout while waiting to sample to be loaded"
        with gevent.Timeout(600, RuntimeError(err_msg)):
            while not load_task.ready():
                loaded_sample = tuple(
                    self._execute_cmd_exporter(
                        "getMountedSamplePosition", attribute=True
                    )
                )

                _use_custom = HWR.beamline.diffractometer.get_property(
                    "use_custom_phase_script", False
                )

                if loaded_sample == (
                    sample.get_cell_no(),
                    sample.get_basket_no(),
                    sample.get_vial_no(),
                ):
                    if _use_custom:
                        logging.getLogger("user_level_log").info(
                            "Sample is Loaded from EMBLFlexHCD.py "
                        )
                        HWR.beamline.diffractometer.wait_status_ready(100)
                        HWR.beamline.diffractometer.run_custom_script(
                            "ChangePhase_centring"
                        )
                        HWR.beamline.diffractometer.run_custom_script(
                            "sample_centering", wait=False
                        )
                    break

                gevent.sleep(1)

        with gevent.Timeout(600, RuntimeError(err_msg)):
            while True:
                is_safe = self._execute_cmd_exporter("getRobotIsSafe", attribute=True)

                if is_safe:
                    break

                gevent.sleep(2)

        for msg in self.get_robot_exceptions():
            if msg is not None:
                self.log.error(msg)
                logging.getLogger("user_level_log").error(msg)

        return self._set_loaded_sample_and_prepare(loaded_sample, previous_sample)

    def _do_unload(self, sample=None):
        # We wait for the sample changer if it is already doing something,
        # like defreezing.
        # wait for 10 minutes then timeout !
        self._wait_ready(600)

        self._execute_cmd_exporter(
            "unloadSample",
            sample.get_cell_no(),
            sample.get_basket_no(),
            sample.get_vial_no(),
            command=True,
        )

        loaded_sample = tuple(
            self._execute_cmd_exporter("getMountedSamplePosition", attribute=True)
        )

        for msg in self.get_robot_exceptions():
            if msg is not None:
                self.log.error(msg)
                logging.getLogger("user_level_log").error(msg)

        if loaded_sample == (-1, -1, -1):
            self._reset_loaded_sample()

            if self.controller:
                self.controller.hutch_actions(release_interlock=True)

            return True

        return False

    def _do_abort(self):
        self._execute_cmd_exporter("abort", command=True)

    def _do_trash(self):
        self.prepare_load()
        self._execute_cmd_exporter("trashMountedSample", command=True)
        self._reset_loaded_sample()

    def _do_reset(self):
        self._execute_cmd_exporter("homeClear", command=True)

    def clear_basket_info(self, basket):
        return self._reset_basket_info(basket)

    def _reset_basket_info(self, basket):
        pass

    def clear_cell_info(self, cell):
        return self._reset_cell_info(cell)

    def _reset_cell_info(self, cell):
        pass

    def _update_state(self):
        try:
            state = self._read_state()
            status = self._execute_cmd_exporter("getStatus", attribute=True)
        except Exception:
            state = SampleChangerState.Unknown
            status = "Unknown"

        self._set_state(state, status)

    def is_sequencer_ready(self):
        if self.prepareLoad:
            cmdobj = self.get_command_object
            return all(
                [cmd.isSpecReady() for cmd in (cmdobj("moveToLoadingPosition"),)]
            )
        return True

    def _read_state(self):
        state = self._execute_cmd_exporter("getState", attribute=True).upper()

        state_converter = {
            "ALARM": SampleChangerState.Alarm,
            "FAULT": SampleChangerState.Fault,
            "RUNNING": SampleChangerState.Moving,
            "READY": SampleChangerState.Ready,
            "STANDBY": SampleChangerState.Ready,
        }

        return state_converter.get(state, SampleChangerState.Unknown)

    def _is_device_busy(self, state=None):
        if state is None:
            state = self._read_state()
        return state not in (
            SampleChangerState.Ready,
            SampleChangerState.Loaded,
            SampleChangerState.Alarm,
            SampleChangerState.Disabled,
            SampleChangerState.Fault,
            SampleChangerState.StandBy,
        )

    def _is_device_ready(self):
        state = self._read_state()
        return state in (SampleChangerState.Ready, SampleChangerState.Charging)

    def _wait_device_ready(self, timeout=None):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self._is_device_ready():
                gevent.sleep(0.01)

    def _update_selection(self):
        sample_cell, sample_puck, sample = self._execute_cmd_exporter(
            "getMountedSamplePosition", attribute=True
        )

        cell = sample_cell

        for cmp in self.get_components():
            idx = cmp.get_index()
            if cell == idx + 1:
                self._set_selected_component(cmp)
                break

        # find sample
        for samp in self.get_sample_list():
            if samp.get_coords() == (sample_cell, sample_puck, sample):
                self._set_loaded_sample(samp)
                self._set_selected_sample(samp)
            else:
                samp._set_loaded(False)

    def prepare_hutch(self, **kwargs):
        return



"""
Combine setting the back light IN with moving the beamstop out.
This is used to set the back light in/out faster, than using
the phase CENTRING - the beamstop motor moves only if needed.
Example xml configuration:

.. code-block:: xml
 <object class="MicrodiffLightInOut">
   <username>Back Light</username>
   <exporter_address>wid30bmd:9001</exporter_address>
   <value_channel_name>BackLightIsOn</value_channel_name>
   <values>{"IN": True, "OUT": False}</values>
   <state_channel_name>HardwareState</state_channel_name>
 </object>
"""

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ExporterNState import ExporterNState

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicrodiffLightInOut(ExporterNState):
    """MicrodiffAperture class"""

    def _set_value(self, value):
        """Set device to value
        Args:
            value (str, int, float or enum): Value to be set.
        """
        # first move the beamstop in our out
        beamstop = HWR.beamline.diffractometer.beamstop
        if value == self.VALUES.IN and beamstop.get_value() == beamstop.VALUES.IN:
            beamstop.set_value(beamstop.VALUES.OUT)
        super()._set_value(value)



import json
import logging
from json.decoder import JSONDecodeError
from typing import (
    Dict,
    List,
    Optional,
)
from urllib.parse import urljoin

import requests
from suds import WebFault

from mxcubecore.HardwareObjects.abstract.ISPyBDataAdapter import ISPyBDataAdapter
from mxcubecore.HardwareObjects.UserTypeISPyBLims import UserTypeISPyBLims
from mxcubecore.model.lims_session import (
    LimsSessionManager,
    Session,
)

log = logging.getLogger("ispyb_client")

LAZY_SESSION_PREFIX = "lazy"


def _get_lazy_session_id(proposal: Dict) -> str:
    prop_id = proposal["proposalId"]
    return f"{LAZY_SESSION_PREFIX}{prop_id}"


def _is_lazy_session_id(session_id: str) -> bool:
    return session_id.startswith(LAZY_SESSION_PREFIX)


def _check_ispyb_error_message(response):
    def _expected_ispyb_err_msg(error_msg):
        import re

        match = re.match("^JBAS011843: Failed instantiate.*ldap.*ispyb", error_msg)
        return match is not None

    #
    # check that we got the 'expected' error message on invalid credentials,
    # otherwise log the error message, so we don't swallow new error messages
    #
    if _expected_ispyb_err_msg(response.text):
        # all is fine
        return

    log.warning(
        "unexpected response from ISPyB\n"
        + f"{response.status_code} {response.reason}\n{response.text}"
    )


class ISPyBRestClient:
    def __init__(self, rest_root: str):
        self._rest_root = rest_root

    def authenticate(self, user_name: str, password: str):
        """
        authenticate with REST services

        Args:
            user_name: Username
            password: Password
        """
        token = None

        auth_url = urljoin(self._rest_root, "authenticate?site=MAXIV")
        response = requests.post(
            auth_url, data={"login": user_name, "password": password}
        )

        try:
            # if authentication is successful, we will get
            # JSON response containing an auth token
            token = response.json().get("token", None)
        except JSONDecodeError:
            # on invalid credentials, some ISPyB systems will reply with
            # an internal error message, as plain text
            _check_ispyb_error_message(response)

        if token is None:
            # we failed to obtain the auth token, thus we failed to authenticate
            raise Exception("invalid credentials")


def _create_session_object(proposal, session_id: str, beamline_name: str) -> Session:
    return Session(
        proposal_id=proposal["proposalId"],
        code=proposal["code"],
        number=proposal["number"],
        session_id=session_id,
        beamline_name=beamline_name,
        title=proposal["title"],
        #
        # At MAXIV we don't care if a session is scheduled
        # or not, mark all sessions as scheduled.
        #
        is_scheduled_time=True,
        is_scheduled_beamline=True,
    )


class CustomISPyBDataAdapter(ISPyBDataAdapter):
    """
    Extend the standard ISPyB data adapter with MAXIV specific logic of how to
    deal with proposal sessions.
    """

    def _get_proposals(self, username: str):
        proposals = json.loads(
            self._shipping.service.findProposalsByLoginName(username)
        )

        for proposal in proposals:
            if proposal["type"].upper() not in ["MX", "MB"]:
                continue
            if proposal.get("state", "Open") != "Open":
                continue

            yield proposal

    def _get_sessions(self, username: str, beamline_name: str) -> List[Session]:
        def list_sessions():
            for proposal in self._get_proposals(username):
                sessions = self._collection.service.findSessionsByProposalAndBeamLine(
                    proposal["code"], proposal["number"], beamline_name
                )

                for sesssion in sessions:
                    yield _create_session_object(
                        proposal, sesssion["sessionId"], beamline_name
                    )

                #
                # A hack to lazily create new sessions.
                #
                # At MAXIV we don't schedule sessions for proposals ahead of time. Instead, we
                # lazily create them as needed.
                #
                # If a proposal does not contain any active session, create a Session object
                # with a special session ID.
                #
                # If user selects such a session, then we will ask ISPyB to create this session.
                #
                if len(sessions) == 0:
                    yield _create_session_object(
                        proposal, _get_lazy_session_id(proposal), beamline_name
                    )

        return sorted(list_sessions(), key=lambda s: f"{s.code}{s.number}")

    def get_sessions_by_username(
        self, username: str, beamline_name: str
    ) -> LimsSessionManager:
        try:
            sessions = list(self._get_sessions(username, beamline_name))
            return LimsSessionManager(sessions=sessions)
        except WebFault as e:
            log.exception(e.message)


class ISPyBLims(UserTypeISPyBLims):
    def init(self):
        super().init()

        self._rest_client = ISPyBRestClient(self.get_property("rest_root"))

    def _create_data_adapter(self) -> ISPyBDataAdapter:
        return CustomISPyBDataAdapter(
            self.ws_root.strip(),
            self.proxy,
            self.ws_username,
            self.ws_password,
            self.beamline_name,
        )

    def ispyb_login(self, user_name: str, password: str):
        try:
            self._rest_client.authenticate(user_name, password)
            return True, None
        except Exception as ex:
            return False, str(ex)

    def set_active_session_by_id(self, session_id: str) -> Session:
        """
        Sets session with session_id to active session

        Args:
            session_id: session id
        """

        def find_session() -> Optional[Session]:
            for session in self.session_manager.sessions:
                if session.session_id == session_id:
                    self.session_manager.active_session = session

                    return session

            # session not found
            return None

        def replace_lazy(sessions: List[Session], new_session: Session):
            def gen():
                for session in sessions:
                    if session.session_id == session_id:
                        yield new_session
                    else:
                        yield session

            return list(gen())

        session = find_session()
        if session is None:
            raise Exception(f"no session with ID {session_id} found")

        #
        # user selected a session that does not exist yet,
        # ask ISPyB to create it
        #
        if _is_lazy_session_id(session_id):
            session = self.adapter.create_session(
                session.proposal_id, session.beamline_name
            )
            # replace the old lazy-session object,
            # with the new proper-session object
            self.session_manager.sessions = replace_lazy(
                self.session_manager.sessions, session
            )

        return session

    def get_full_user_name(self) -> str:
        person = self.adapter.get_person_by_username(self.user_name)

        given_name = person["givenName"]
        family_name = person["familyName"]

        return f"{given_name} {family_name}"



from mxcubecore.HardwareObjects import SpecMotor


class SpecMotorWPositions(SpecMotor.SpecMotor):
    def init(self):
        self.predefinedPositions = {}
        self.predefinedPositionsNamesList = []
        self.delta = self.get_property("delta") or 0

        try:
            positions = self["positions"]
        except Exception:
            self.log.error("%s does not define positions.", str(self.name()))
        else:
            for definedPosition in positions:
                positionUsername = definedPosition.get_property("username")

                try:
                    offset = float(definedPosition.get_property("offset"))
                except Exception:
                    self.log.warning(
                        "%s, ignoring position %s: invalid offset.",
                        str(self.name()),
                        positionUsername,
                    )
                else:
                    self.predefinedPositions[positionUsername] = offset

            self.sortPredefinedPositionsList()

    def connect_notify(self, signal):
        SpecMotor.SpecMotor.connect_notify.__func__(self, signal)

        if signal == "predefinedPositionChanged":
            positionName = self.get_current_position_name()

            try:
                pos = self.predefinedPositions[positionName]
            except KeyError:
                self.emit(signal, ("", None))
            else:
                self.emit(signal, (positionName, pos))
        elif signal == "stateChanged":
            self.emit(signal, (self.get_state(),))

    def sortPredefinedPositionsList(self):
        self.predefinedPositionsNamesList = list(self.predefinedPositions.keys())
        self.predefinedPositionsNamesList.sort(
            lambda x, y: int(
                round(self.predefinedPositions[x] - self.predefinedPositions[y])
            )
        )

    def motorMoveDone(self, channelValue):
        SpecMotor.SpecMotor.motorMoveDone.__func__(self, channelValue)

        pos = self.get_value()
        self.log.debug("current pos=%s", pos)

        for positionName in self.predefinedPositions:
            if (
                self.predefinedPositions[positionName] >= pos - self.delta
                and self.predefinedPositions[positionName] <= pos + self.delta
            ):
                self.emit("predefinedPositionChanged", (positionName, pos))
                return

        self.emit("predefinedPositionChanged", ("", None))

    def get_predefined_positions_list(self):
        return self.predefinedPositionsNamesList

    def moveToPosition(self, positionName):
        try:
            self.move(self.predefinedPositions[positionName])
        except Exception:
            self.log.exception(
                "Cannot move motor %s: invalid position name.", str(self.username)
            )

    def get_current_position_name(self):
        if (
            not self.motorIsMoving()
        ):  # self.is_ready() and self.get_state() == self.READY:
            for positionName in self.predefinedPositions:
                if (
                    self.predefinedPositions[positionName]
                    >= self.get_value() - self.delta
                    and self.predefinedPositions[positionName]
                    <= self.get_value() + self.delta
                ):
                    return positionName
        return ""

    def setNewPredefinedPosition(self, positionName, positionOffset):
        try:
            self.predefinedPositions[str(positionName)] = float(positionOffset)
            self.sortPredefinedPositionsList()
        except Exception:
            self.log.exception("Cannot set new predefined position")



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Motor class.
Defines the MotorStates enum, get/set velocity, home and set_value_relative
methods.
Emits signals valueChanged and limitsChanged.
"""

import abc
from enum import (
    Enum,
    unique,
)

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractActuator import AbstractActuator

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@unique
class MotorStates(Enum):
    """Motor states definitions."""

    HOME = HardwareObjectState.READY, 5
    LOWLIMIT = HardwareObjectState.READY, 6
    HIGHLIMIT = HardwareObjectState.READY, 7
    MOVING = HardwareObjectState.BUSY, 8


class AbstractMotor(AbstractActuator):
    """Abstract motor class"""

    __metaclass__ = abc.ABCMeta
    unit = None

    SPECIFIC_STATES = MotorStates

    def __init__(self, name):
        super().__init__(name)
        self._velocity = None
        self._tolerance = None

    def init(self):
        """Initialise tolerance property"""
        super().init()
        self._tolerance = self.get_property("tolerance") or 1e-3

    def get_velocity(self):
        """Read motor velocity.
        Returns:
            (float): velocity [unit/s]
        """
        return self._velocity

    def set_velocity(self, velocity):
        """Set the motor velocity
        Args:
            velocity (float): target velocity
        """
        self._velocity = velocity

    def set_value_relative(self, relative_value, timeout=0):
        """
        Set actuator to relative to the current value
        Args:
            value (float): target value
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait;
                             if timeout is None: wait forever.
        """
        self.set_value(self.get_value() + relative_value, timeout)

    def home(self, timeout=None):
        """Homing procedure.
        Args:
            timeout (float): Timeout [s].
        """
        raise NotImplementedError

    def update_value(self, value=None):
        """Check if the value has changed. Emits signal valueChanged.
        Args:
            value (float): value
        """
        if value is None:
            value = self.get_value()

        if self._nominal_value is None:
            if value is None:
                return

        elif value is not None and self._tolerance:
            if abs(value - self._nominal_value) <= self._tolerance:
                return

        self._nominal_value = value
        self.emit("valueChanged", (value,))



"""
A client for ISPyB Webservices.
"""

from pyicat_plus.client.main import IcatClient

from mxcubecore.HardwareObjects.ICATLIMS import ICATLIMS


class LNLSLIMS(ICATLIMS):
    def init(self):
        self.url = self.get_property("ws_root")
        self.ingesters = self.get_property("queue_urls")
        self.investigations = []
        self.samples = []

        self.icatClient = IcatClient(
            icatplus_restricted_url="https://icat-plus.cnpem.br"
        )

    def is_single_session_available(self):
        """
        True if there is no active session and there is
        a single session available
        """
        return (
            self.session_manager.active_session is None
            and len(self.session_manager.sessions) == 1
        )

    def _create_icat_session(self, user_name: str, password: str):
        self.icat_session = self.icatClient.do_log_in(
            username=user_name, password=password, plugin="oidc"
        )

    def login(self, user_name, token, is_local_host):
        self.is_local_host = is_local_host
        session_manager, lims_username, sessions = super().login(
            user_name, token, self.session_manager
        )
        self.session_manager = session_manager
        self.add_user_and_shared_sessions(lims_username, sessions)
        if self.is_single_session_available():
            single_session = self.session_manager.sessions[0]
            self.set_active_session_by_id(single_session.session_id)

        return session_manager



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import base64
import time

import numpy as np
import urllib2

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice
from mxcubecore.utils import qt_import

"""
Hardware objects allows to access Axis camera jpg frames via direct http requests.

Example xml:
<object class="AxisCamera">
   <interval>1000</interval>
   <address>ADDRESS_OF_THE_CAMERA/axis-cgi/jpg/image.cgi</address>
   <user>USER</user>
   <password>PASSWORD</password>
   <image_size>(600, 480)</image_size>
</object>
"""


class QtAxisCamera(AbstractVideoDevice):
    def __init__(self, name):
        AbstractVideoDevice.__init__(self, name)

    def init(self):
        self.image_dimensions = [352, 288]
        self.request = urllib2.Request(self.get_property("address"))
        base64string = base64.b64encode(
            "%s:%s" % (self.get_property("user"), self.get_property("password"))
        )
        self.request.add_header("Authorization", "Basic %s" % base64string)
        self.qpixmap = QPixmap()
        self.set_is_ready(True)

        AbstractVideoDevice.init(self)

        self.set_video_live(False)

    def get_new_image(self):
        result = urllib2.urlopen(self.request)
        self.qpixmap.loadFromData(result.read(), "JPG")
        self.emit("imageReceived", self.qpixmap)

    def save_snapshot(self, filename, image_type="PNG"):
        qimage = qt_import.QImage(self.image)
        qimage.save(filename, image_type)

    def get_snapshot(self, bw=None, return_as_array=None):
        qimage = self.qimage
        if return_as_array:
            qimage = qimage.convertToFormat(4)
            ptr = qimage.bits()
            ptr.setsize(qimage.byteCount())

            image_array = np.array(ptr).reshape(qimage.height(), qimage.width(), 4)
            if bw:
                return np.dot(image_array[..., :3], [0.299, 0.587, 0.144])
            else:
                return image_array
        else:
            if bw:
                return qimage.convertToFormat(qt_import.QImage.Format_Mono)
            else:
                return qimage

    def get_raw_image_size(self):
        return list(self.image_dimensions)

    def do_image_polling(self, sleep_time):
        while True:
            if self.get_video_live():
                self.get_new_image()
            time.sleep(sleep_time)



import logging
import math
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import (
    HardwareObject,
    HardwareObjectState,
)
from mxcubecore.HardwareObjects import (
    MiniDiff,
    sample_centring,
)

MICRODIFF = None

EXPORTER_TO_HWOBJ_STATE = {
    "Fault": HardwareObjectState.FAULT,
    "Ready": HardwareObjectState.READY,
    "Running": HardwareObjectState.BUSY,
    "Moving": HardwareObjectState.BUSY,
    "Busy": HardwareObjectState.BUSY,
    "Unknown": HardwareObjectState.BUSY,
    "Offline": HardwareObjectState.OFF,
}


class Microdiff(MiniDiff.MiniDiff):
    def init(self):
        global MICRODIFF
        MICRODIFF = self
        self.phiMotor = self.get_object_by_role("phi")
        self.exporter_addr = self.get_property("exporter_address")

        self.update_state(HardwareObjectState.READY)

        self.x_calib = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "x_calib",
            },
            "CoaxCamScaleX",
        )
        self.y_calib = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "y_calib",
            },
            "CoaxCamScaleY",
        )
        self.moveMultipleMotors = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "move_multiple_motors",
            },
            "SyncMoveMotors",
        )
        self.head_type = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "head_type",
            },
            "HeadType",
        )
        self.kappa_channel = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "kappa_enable",
            },
            "KappaIsEnabled",
        )
        self.phases = {
            "Centring": 1,
            "BeamLocation": 2,
            "DataCollection": 3,
            "Transfer": 4,
        }
        self.move_phase = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "move_to_phase",
            },
            "startSetPhase",
        )
        self.readPhase = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "read_phase",
            },
            "CurrentPhase",
        )

        self.state = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "state",
            },
            "State",
        )

        self.scanLimits = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "scan_limits",
            },
            "getOmegaMotorDynamicScanLimits",
        )
        if self.get_property("use_hwstate"):
            self.hwstate_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "hwstate",
                },
                "HardwareState",
            )
        else:
            self.hwstate_attr = None
        self.swstate_attr = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "swstate",
            },
            "State",
        )
        self.nb_frames = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "nbframes",
            },
            "ScanNumberOfFrames",
        )

        # raster scan attributes
        self.scan_range = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "scan_range",
            },
            "ScanRange",
        )
        self.scan_exposure_time = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "exposure_time",
            },
            "ScanExposureTime",
        )
        self.scan_start_angle = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_angle",
            },
            "ScanStartAngle",
        )
        self.scan_detector_gate_pulse_enabled = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "detector_gate_pulse_enabled",
            },
            "DetectorGatePulseEnabled",
        )
        self.scan_detector_gate_pulse_readout_time = self.add_channel(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "detector_gate_pulse_readout_time",
            },
            "DetectorGatePulseReadoutTime",
        )

        self.abort_cmd = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "abort",
            },
            "abort",
        )

        self._move_sync_motors = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "move_sync_motors",
            },
            "startSimultaneousMoveMotors",
        )

        self.beam_position_horizontal = self.add_channel(
            {"type": "exporter", "exporter_address": self.exporter_addr, "name": "bph"},
            "BeamPositionHorizontal",
        )

        self.beam_position_vertical = self.add_channel(
            {"type": "exporter", "exporter_address": self.exporter_addr, "name": "bpv"},
            "BeamPositionVertical",
        )

        self.save_centring_positions = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "save_centring_positions",
            },
            "saveCentringPositions",
        )

        self.set_room_temperature_mode = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "set_room_temperature_mode",
            },
            "setRoomTemperatureMode",
        )

        self.prepare_ssx_grid_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "prepare_ssx_grid_scan",
            },
            "prepareSSXGridScan",
        )

        self.start_ssx_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_scan",
            },
            "startSSXGridScan",
        )

        self.start_still_ssx_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_still_ssx_scan",
            },
            "startSSXStillScan",
        )

        self.prepare_ssx_grid_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "prepare_ssx_grid_scan",
            },
            "prepareSSXGridScan",
        )

        self.get_ssx_scan_method = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "get_ssx_scan_method",
            },
            "getSSXScanMethod",
        )

        self.get_ssx_delay = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "get_ssx_delay",
            },
            "getSSXDeltaT",
        )

        self.prepare_ssx_line_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "prepare_ssx_line_scan",
            },
            "prepareSSXLineScan",
        )

        self.start_ssx_line_scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "startSSXLineScan",
        )

        self.get_last_task_info = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "get_last_task_info",
            },
            "getLastTaskInfo",
        )

        #
        # Args: doublex, doubley, doublez
        # Retruns: 1 if success else other than 1
        #
        self.add_ssx_chip_calibration_fiducial = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "addSSXChipCalibrationFiducial",
        )

        self.add_ssx_chip_calibration_point = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "addSSXChipCalibrationPoint",
        )

        self.add_ssx_chip_calibration_point = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "addSSXChipCalibrationPoint",
        )

        self.add_ssx_chip_calibration_point = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "resetSSXChipCalibration",
        )

        self.get_ssx_chip_calibration_state = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "getSSXChipCalibrationState",
        )

        self.get_ssx_block_calibration_state = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "getSSXBlockCalibrationState",
        )

        self.start_ssx_all_block_calibration = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "startSSXAllBlockCalibration",
        )

        self.reset_ssx_all_block_calibration = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "resetSSXAllBlockCalibration",
        )

        self.ir_auto_focus = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_line_scan",
            },
            "doSSXAutoFocus",
        )

        self.start_ssx_all_block_calibration = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_ssx_all_block_calibration",
            },
            "startSSXAllBlockCalibration",
        )

        MiniDiff.MiniDiff.init(self)
        self.centringPhiy.direction = -1
        self.MOTOR_TO_EXPORTER_NAME = self.getMotorToExporterNames()
        self.move_to_coord = self.move_to_beam

        self.centringVertical = self.get_object_by_role("sample_vertical")
        self.centringFocus = self.get_object_by_role("sample_focus")

        self.saved_motor_position = {}

        self.frontLight = self.get_object_by_role("FrontLight")
        self.backLight = self.get_object_by_role("BackLight")

        self.wait_ready = self._wait_ready
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(None)

        self.readPhase.connect_signal("update", self._update_value)
        self.state.connect_signal("update", self._update_state)

        HardwareObject.init(self)
        self.handle_detector_cover = self.get_object_by_role("handle_detcover")

    def use_position_for_calibration(self, data):
        for _d in data:
            if len(data[_d]) > 0:
                x, y, z = data[_d]
                logging.getLogger("HWR").info(
                    f"Setting fiducial {_d} to: ({x}, {y}, {z})"
                )
                self.add_ssx_chip_calibration_fiducial(x, y, z)

        self.start_ssx_all_block_calibration()

        res = 1
        return res == 1

    def _update_value(self, value=None):
        if value is None:
            value = self.get_current_phase()
        self.emit("phaseChanged", (value))

    def _update_state(self, value):
        self.update_state(
            EXPORTER_TO_HWOBJ_STATE.get(value, HardwareObjectState.UNKNOWN)
        )

    def abort(self):
        self.abort_cmd()
        return True

    def getMotorToExporterNames(self):
        MOTOR_TO_EXPORTER_NAME = {
            "focus": self.focusMotor.get_property("actuator_name"),
            "phi": self.phiMotor.get_property("actuator_name"),
            "phiy": self.phiyMotor.get_property("actuator_name"),
            "phiz": self.phizMotor.get_property("actuator_name"),
            "sampx": self.sampleXMotor.get_property("actuator_name"),
            "sampy": self.sampleYMotor.get_property("actuator_name"),
            "zoom": "Zoom",
        }
        if self.in_kappa_mode:
            MOTOR_TO_EXPORTER_NAME.update(
                {"kappa": self.kappaMotor.get_property("actuator_name")}
            )
            MOTOR_TO_EXPORTER_NAME.update(
                {"kappa_phi": self.kappaPhiMotor.get_property("actuator_name")}
            )

        return MOTOR_TO_EXPORTER_NAME

    def save_current_motor_position(self):
        motor_pos_dict = {
            "focus": self.focusMotor.get_value(),
            "phiy": self.phiyMotor.get_value(),
            "phiz": self.phizMotor.get_value(),
            "centring_focus": self.centringFocus.get_value(),
            "centring_vertical": self.centringVertical.get_value(),
        }
        self.saved_motor_position = motor_pos_dict

    def getCalibrationData(self, offset):
        return (1.0 / self.x_calib.get_value(), 1.0 / self.y_calib.get_value())

    def emitCentringSuccessful(self):
        # check first if all the motors have stopped
        self._wait_ready(30)

        # do normal stuff
        return MiniDiff.MiniDiff.emitCentringSuccessful(self)

    def _ready(self):
        if self.hwstate_attr:
            if (
                self.hwstate_attr.get_value() == "Ready"
                and self.swstate_attr.get_value() == "Ready"
            ):
                return True
        else:
            if self.swstate_attr.get_value() == "Ready":
                return True
        return False

    def _wait_ready(self, timeout=None):
        # None means infinite timeout
        # <=0 means default timeout
        if timeout is not None and timeout <= 0:
            self.log.warning("DEBUG: Strange timeout value passed %s" % str(timeout))
            timeout = 30
        with gevent.Timeout(
            timeout, RuntimeError("Timeout waiting for diffractometer to be ready")
        ):
            while not self._ready():
                time.sleep(0.5)

    def open_detector_cover(self, timeout=10):
        use = True
        if self.handle_detector_cover:
            try:
                use = self.handle_detector_cover.get_value().value
            except AttributeError:
                use = False

        logging.getLogger("user_level_log").info("Use detector cover open: %s" % use)

        if use:
            try:
                detcover = self.get_object_by_role("controller").detcover

                if detcover.state == "IN":
                    detcover.set_out(timeout)
            except AttributeError:
                logging.getLogger("user_level_log").exception(
                    "No detector cover configured"
                )

    def close_detector_cover(self, timeout=10):
        use = True
        if self.handle_detector_cover:
            try:
                use = self.handle_detector_cover.get_value().value
            except AttributeError:
                use = False

        logging.getLogger("user_level_log").info("Use detector cover close: %s" % use)

        if use:
            try:
                detcover = self.get_object_by_role("controller").detcover

                if detcover.state == "OUT":
                    detcover.set_in(timeout)
            except AttributeError:
                logging.getLogger("user_level_log").exception(
                    "No detector cover configured"
                )

    def phase_prepare(self, phase):
        if phase == "Centring":
            try:
                diffr = self.get_object_by_role("controller").diffractometer
                diffr.prepare("centre")
            except Exception:
                self.log.exception("Cannot prepare centring")

    def set_light_in(self):
        """Set the backlight in - used by the XMLRPC calls"""
        self.log.info("Moving backlight in")
        light_hwobj = self.get_object_by_role("BackLightSwitch")
        light_hwobj.set_value(light_hwobj.VALUES.IN)
        self._wait_ready(20)

    def set_light_out(self):
        """Set the backlight out - used by the XMLRPC calls"""
        self.log.info("Moving backlight out")
        light_hwobj = self.get_object_by_role("BackLightSwitch")
        light_hwobj.set_value(light_hwobj.VALUES.OUT)
        self._wait_ready(20)

    def set_phase(self, phase, wait=False, timeout=None):
        """Set the phase"""
        _use_custom = self.get_property("use_custom_phase_script", False)
        if not self._ready():
            self.log.exception("MD not ready - phase not set.")
            return

        current_phase = self.get_current_phase()

        msg = f"Current phase is {current_phase} and moving to {phase}"
        logging.getLogger("user_level_log").info(msg)

        if phase in self.phases:
            if phase in ["BeamLocation", "Transfer"]:
                self.close_detector_cover()
            self.phase_prepare(phase)

            if _use_custom and not self.in_plate_mode:
                script = "ChangePhase_" + phase.lower()
                msg = f"Changing phase to {phase}, using pmac script"
                logging.getLogger("user_level_log").info(msg)
                self.run_script(script)
                self._wait_ready(600)
            else:
                self.move_phase(phase)
            if wait:
                timeout = timeout or 40
                self._wait_ready(timeout)

    def get_current_phase(self):
        return self.readPhase.get_value()

    def get_phase_list(self):
        return list(self.phases.keys())

    def move_sync_motors(self, motors_dict, wait=False, timeout=None):
        in_kappa_mode = self.in_kappa_mode
        argin = ""
        # print "start moving motors =============", time.time()
        if wait:
            self._wait_ready()
        for motor in motors_dict.keys():
            position = motors_dict[motor]
            if position is None:
                continue

            if not in_kappa_mode and motor in ("kappa", "kappa_phi"):
                continue

            name = self.MOTOR_TO_EXPORTER_NAME[motor]
            argin += "%s=%0.3f;" % (name, position)
        if not argin:
            return

        self._move_sync_motors(argin)

        if wait:
            time.sleep(0.1)
            self._wait_ready()
        # print "end moving motors =============", time.time()

    def oscilScan(self, start, end, exptime, number_of_images, wait=False):
        if self.in_plate_mode:
            scan_speed = math.fabs(end - start) / exptime
            low_lim, hi_lim = map(float, self.scanLimits(scan_speed))
            if start < low_lim:
                raise ValueError("Scan start below the allowed value %f" % low_lim)
            elif end > hi_lim:
                raise ValueError("Scan end above the allowed value %f" % hi_lim)

        dead_time = HWR.beamline.detector.get_deadtime()

        self.scan_detector_gate_pulse_enabled.set_value(True)
        self.scan_detector_gate_pulse_readout_time.set_value(dead_time * 1000)

        if self.get_property("md_set_number_of_frames", False):
            self.nb_frames.set_value(number_of_images)
        else:
            self.nb_frames.set_value(1)

        scan_params = "1\t%0.3f\t%0.3f\t%0.4f\t1" % (start, (end - start), exptime)
        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_scan",
            },
            "startScanEx",
        )
        scan(scan_params)
        print("oscil scan started at ----------->", time.time())
        if wait:
            self._wait_ready(40 * 60)  # Timeout of 40 min

    def oscilScan4d(
        self, start, end, exptime, number_of_images, motors_pos, wait=False
    ):
        if self.in_plate_mode:
            scan_speed = math.fabs(end - start) / exptime
            low_lim, hi_lim = map(float, self.scanLimits(scan_speed))
            if start < low_lim:
                raise ValueError("Scan start below the allowed value %f" % low_lim)
            elif end > hi_lim:
                raise ValueError("Scan end above the allowed value %f" % hi_lim)

        if self.get_property("md_set_number_of_frames", False):
            self.nb_frames.set_value(number_of_images)
        else:
            self.nb_frames.set_value(1)

        scan_params = "%0.3f\t%0.3f\t%f\t" % (start, (end - start), exptime)
        scan_params += "%0.3f\t" % motors_pos["1"]["phiy"]
        scan_params += "%0.3f\t" % motors_pos["1"]["phiz"]
        scan_params += "%0.3f\t" % motors_pos["1"]["sampx"]
        scan_params += "%0.3f\t" % motors_pos["1"]["sampy"]
        scan_params += "%0.3f\t" % motors_pos["2"]["phiy"]
        scan_params += "%0.3f\t" % motors_pos["2"]["phiz"]
        scan_params += "%0.3f\t" % motors_pos["2"]["sampx"]
        scan_params += "%0.3f" % motors_pos["2"]["sampy"]

        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_scan4d",
            },
            "startScan4DEx",
        )

        scan(scan_params)

        print("helical scan started at ----------->", time.time())
        if wait:
            self._wait_ready(20 * 60)  # timeout of 15 min
            print("finished at ---------->", time.time())

    def oscilScanMesh(
        self,
        start,
        end,
        exptime,
        dead_time,
        mesh_num_lines,
        mesh_total_nb_frames,
        mesh_center,
        mesh_range,
        wait=False,
    ):
        self.scan_detector_gate_pulse_enabled.set_value(True)
        dead_time = HWR.beamline.detector.get_deadtime()
        self.scan_detector_gate_pulse_readout_time.set_value(dead_time * 1000)
        self.move_motors(mesh_center.as_dict())
        positions = self.get_positions()

        params = "%0.3f\t" % (end - start)
        params += "%0.3f\t" % -mesh_range["horizontal_range"]
        params += "%0.3f\t" % mesh_range["vertical_range"]
        params += "%0.3f\t" % start
        params += "%0.3f\t" % positions["phiy"]
        params += "%0.3f\t" % positions["phiz"]
        params += "%0.3f\t" % positions["sampx"]
        params += "%0.3f\t" % positions["sampy"]
        params += "%d\t" % mesh_num_lines
        params += "%d\t" % (mesh_total_nb_frames / mesh_num_lines)
        params += "%0.6f\t" % (exptime / mesh_num_lines)
        params += "%r\t" % True
        params += "%r\t" % True
        params += "%r\t" % self.get_property("use_fast_mesh", True)

        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_raster_scan",
            },
            "startRasterScanEx",
        )

        self._wait_ready(900)  # timeout of 30 min

        scan(params)

        if wait:
            # timeout of 30 min
            self._wait_ready(1800)

    def stillScan(self, pulse_duration, pulse_period, pulse_nb, wait=False):
        scan_params = "%0.6f\t%0.6f\t%d" % (pulse_duration, pulse_period, pulse_nb)
        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "start_scan",
            },
            "startStillScan",
        )
        scan(scan_params)
        print("still scan started at ----------->", time.time())
        if wait:
            self._wait_ready(1800)  # timeout of 30 min
            print("finished at ---------->", time.time())

    def characterisation_scan(
        self,
        start,
        scan_range,
        nb_frames,
        exptime,
        nb_scans,
        angle,
        wait=False,
    ):
        """Do N scans continuously.
        Args:
            start (float): Position of omega for the first scan [deg].
            scan_range (float): range for each scan [deg].
            nb_frames (int): Frame numbers for each scan.
            exptime (float): Total exposure time for each scan [s].
            nb_scans (int): How many times a scan to be repeated.
            angle (float): The angle between each scan [deg]. This number,
                           added to the last position of each scan and will
                           be the start position of the consequent scan.
            wait (bool); Wait (True) or no (False) the end of the command.
        """

        if self.in_plate_mode:
            # to see if needed when plates
            return
        scan_params = "%d\t%0.3f\t%0.3f\t" % (nb_frames, start, scan_range)
        scan_params += "%0.3f\t%d\t%0.3f" % (exptime, nb_scans, angle)

        scan = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "characterisation_scan",
            },
            "startCharacterisationScanEx",
        )

        scan(scan_params)

        print("characterisation scan started at ----------->", time.time())
        if wait:
            self._wait_ready(20 * 60)  # timeout of 15 min
            print("finished at ---------->", time.time())

    @property
    def in_plate_mode(self):
        try:
            return self.head_type.get_value() == "Plate"
        except Exception:
            return False

    @property
    def in_kappa_mode(self):
        return (
            self.head_type.get_value() == "MiniKappa" and self.kappa_channel.get_value()
        )

    def get_motors(self):
        """Get motor_name:Motor dictionary"""
        return {
            "phi": self.phiMotor,
            "focus": self.focusMotor,
            "phiy": self.phiyMotor,
            "phiz": self.phizMotor,
            "sampx": self.sampleXMotor,
            "sampy": self.sampleYMotor,
            "zoom": self.zoomMotor,
            "kappa": self.kappaMotor if self.in_kappa_mode else None,
            "kappa_phi": self.kappaPhiMotor if self.in_kappa_mode else None,
        }

    def get_positions(self):
        pos = {
            "phi": float(self.phiMotor.get_value()),
            "focus": float(self.focusMotor.get_value()),
            "phiy": float(self.phiyMotor.get_value()),
            "phiz": float(self.phizMotor.get_value()),
            "sampx": float(self.sampleXMotor.get_value()),
            "sampy": float(self.sampleYMotor.get_value()),
            "zoom": self.zoomMotor.get_value().value,
            "kappa": (
                float(self.kappaMotor.get_value()) if self.in_kappa_mode else None
            ),
            "kappa_phi": (
                float(self.kappaPhiMotor.get_value()) if self.in_kappa_mode else None
            ),
        }
        return pos

    def move_motors(self, roles_positions_dict):
        self.move_sync_motors(roles_positions_dict, wait=True)

    def move_to_beam(self, x, y):
        if not self.in_plate_mode:
            MiniDiff.MiniDiff.move_to_beam(self, x, y)
        else:
            try:
                beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

                self.centringVertical.set_value_relative(
                    self.centringPhiz.direction
                    * (y - beam_pos_y)
                    / float(self.pixelsPerMmZ)
                )
                self.centringPhiy.set_value_relative(
                    self.centringPhiy.direction
                    * (x - beam_pos_x)
                    / float(self.pixelsPerMmY)
                )

            except Exception:
                logging.getLogger("user_level_log").exception(
                    "Microdiff: could not move to beam, aborting"
                )

    def start_manual_centring(self, sample_info=None):
        self._wait_ready(5)

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

        self.log.info("Starting centring procedure ...")

        if self.in_plate_mode:
            plateTranslation = self.get_object_by_role("plateTranslation")
            cmd_set_plate_vertical = self.add_command(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "plate_vertical",
                },
                "setPlateVertical",
            )
            low_lim, high_lim = self.phiMotor.get_dynamic_limits()
            self.current_centring_procedure = sample_centring.start_plate_1_click(
                {
                    "phi": self.centringPhi,
                    "phiy": self.centringPhiy,
                    "sampx": self.centringSamplex,
                    "sampy": self.centringSampley,
                    "phiz": self.centringVertical,
                    "plateTranslation": plateTranslation,
                },
                self.pixelsPerMmY,
                self.pixelsPerMmZ,
                beam_pos_x,
                beam_pos_y,
                cmd_set_plate_vertical,
                low_lim + 0.5,
                high_lim - 0.5,
            )
        else:
            self.current_centring_procedure = sample_centring.start(
                {
                    "phi": self.centringPhi,
                    "phiy": self.centringPhiy,
                    "sampx": self.centringSamplex,
                    "sampy": self.centringSampley,
                    "phiz": self.centringPhiz,
                },
                self.pixelsPerMmY,
                self.pixelsPerMmZ,
                beam_pos_x,
                beam_pos_y,
                chi_angle=self.chiAngle,
            )

        self.current_centring_procedure.link(self.manualCentringDone)

    def interrupt_and_accept_centring(self):
        """Used when plate. Kills the current 1 click centring infinite loop
        and accepts fake centring - only save the motor positions
        """
        self.current_centring_procedure.kill()
        self.do_centring = False
        self.start_centring_method(self, self.MANUAL3CLICK_MODE)
        self.do_centring = True

    def start_harvester_centring(self, computed_offset: tuple[float]):
        """used when Pin from Harvester"""

        phiy_offset, centringFocus, centringTableVertical = computed_offset

        motor_pos_dict = {
            "kappa": float(
                self["HacentringReferencePosition"].get_property("kappa_ref")
            ),
            "kappa_phi": float(
                self["HacentringReferencePosition"].get_property("phi_ref")
            ),
            "phi": float(self["HacentringReferencePosition"].get_property("omega_ref")),
            "phiy": self.phiyMotor.get_value() + phiy_offset,
        }

        self.move_motors(motor_pos_dict)

        self.centringFocus.set_value_relative(centringFocus, None)

        self.centringVertical.set_value_relative(centringTableVertical, None)

    def getFrontLightLevel(self):
        return self.frontLight.get_value()

    def setFrontLightLevel(self, level):
        return self.frontLight.set_value(level)

    def getBackLightLevel(self):
        return self.backLight.get_value()

    def setBackLightLevel(self, level):
        return self.backLight.set_value(level)

    def get_beam_position(self):
        return (
            self.beam_position_horizontal.get_value(),
            self.beam_position_vertical.get_value(),
        )


def to_float(d):
    for k, v in d.items():
        try:
            d[k] = float(v)
        except Exception:
            pass



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Inherited from AbstracAperture"""

from mxcubecore.HardwareObjects.abstract.AbstractAperture import AbstractAperture

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


DEFAULT_POSITION_LIST = ("BEAM", "OFF", "PARK")


class EMBLAperture(AbstractAperture):
    """Aperture control hwobj uses exporter or Tine channels and commands
    to control aperture position
    """

    def __init__(self, name):
        """Inherited from AbstractAperture"""
        AbstractAperture.__init__(self, name)

        self.chan_diameter_index = None
        self.chan_diameters = None
        self.chan_position = None
        self.chan_state = None

    def init(self):
        """
        Connects to necessary channels
        Returns:
        """
        self._position_list = DEFAULT_POSITION_LIST

        self.chan_diameters = self.get_channel_object("ApertureDiameters")
        if self.chan_diameters:
            self._diameter_size_list = self.chan_diameters.get_value()
        else:
            self._diameter_size_list = (10, 20)

        self.chan_diameter_index = self.get_channel_object(
            "CurrentApertureDiameterIndex"
        )
        if self.chan_diameter_index is not None:
            self._current_diameter_index = self.chan_diameter_index.get_value()
            self.diameter_index_changed(self._current_diameter_index)
            self.chan_diameter_index.connect_signal(
                "update", self.diameter_index_changed
            )
        else:
            self._current_diameter_index = 0

        self.chan_position = self.get_channel_object("AperturePosition")
        if self.chan_position:
            self._current_position_name = self.chan_position.get_value()
            self.current_position_name_changed(self._current_position_name)
            self.chan_position.connect_signal(
                "update", self.current_position_name_changed
            )

        self.chan_state = self.get_channel_object("State")

    def diameter_index_changed(self, diameter_index):
        """Callback when diameter index has been changed"""
        self._current_diameter_index = diameter_index
        self.emit(
            "diameterIndexChanged",
            self._current_diameter_index,
            self._diameter_size_list[self._current_diameter_index] / 1000.0,
        )

    def get_diameter_size(self):
        """
        Returns: diameter size in mm
        """
        return self._diameter_size_list[self._current_diameter_index] / 1000.0

    def current_position_name_changed(self, position):
        """
        Position change callback

        Args:
            position: aperture position (str)
        Returns:
        """
        if position != "UNKNOWN":
            self.set_position_name(position)

    def set_diameter_index(self, diameter_index):
        """
        Sets aperture diameter

        Args:
            diameter_index: diameter index (int)
        Returns:
        """
        self.chan_diameter_index.set_value(diameter_index)

    def set_diameter(self, diameter_size, timeout=None):
        """
        Sets new aperture size

        Args:
            diameter_size: diameter size in microns (int)
            timeout: wait timeout is seconds
        Returns:
        """
        diameter_index = self._diameter_size_list.index(diameter_size)
        self.chan_diameter_index.set_value(diameter_index)
        # if timeout:
        #    while diameter_index != self.chan_diameter_index.get_value():
        #         gevent.sleep(0.1)

    def set_position_index(self, position_index):
        """
        Sets new aperture position

        Args:
            position_index: position index (int)
        Returns:
        """
        self.chan_position.set_value(self._position_list[position_index])

    def set_in(self):
        """
        Sets aperture to the BEAM position

        Returns:
        """
        self.chan_position.set_value("BEAM")

    def set_out(self):
        """Sets aperture to the OUT position

        Returns:
        """
        self.chan_position.set_value("OFF")

    def wait_ready(self, timeout=20):
        """Waits till aperture is ready

        Returns:
        """
        super(EMBLAperture, self).wait_ready(timeout=20)

    def is_out(self):
        """Returns True if aperture is on the beam

        Returns: True if is in, otherwise returns False
        """
        return self._current_position_name != "BEAM"



"""Models the `epics` section of YAML hardware configuration file.

Provides an API to read configured EPICS channels.
"""

from typing import (
    Dict,
    Iterable,
    Optional,
    Tuple,
)

from pydantic import (
    BaseModel,
    RootModel,
)


class Channel(BaseModel):
    """EPICS channel configuration."""

    suffix: Optional[str] = None
    polling_period: Optional[int] = None


class Prefix(BaseModel):
    """Configuration of an EPICS prefix section."""

    channels: Optional[Dict[str, Optional[Channel]]]

    def get_channels(self) -> Iterable[Tuple[str, Channel]]:
        """Get all channels configured for prefix.

        This method will fill in optional configuration properties for a channel.
        """

        if self.channels is None:
            return []

        for channel_name, channel_config in self.channels.items():
            if channel_config is None:
                channel_config = Channel()  # noqa: PLW2901

            if channel_config.suffix is None:
                channel_config.suffix = channel_name

            yield channel_name, channel_config


class EpicsConfig(RootModel[Dict[str, Prefix]]):
    """The 'epics' section of the hardware object's YAML configuration file."""

    def get_prefixes(self) -> Iterable[Tuple[str, Prefix]]:
        """Get all prefixes specified in this section."""
        return list(self.root.items())



from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractNState import (
    AbstractNState,
    BaseValueEnum,
)
from mxcubecore.HardwareObjects.LNLS.EPICS.EPICSActuator import EPICSActuator


class EPICSNState(EPICSActuator, AbstractNState):
    """
    This class manages devices that exist in a discrete number of states
    and provides an interface between those states (configured in the yaml
    files) and the EPICS PVs. It ensures that hardware states are treated as Enums.

    YAML Example
    ------------

    %YAML 1.2
    ---
    class: LNLS.EPICS.EPICSNState.EPICSNState
    epics:
      "MNC:B:PB03:PV_ACTIVATE_BACKLIGHT":
        channels:
          rbv:
            suffix: ':RBV'
            polling_period: 200
          val:
            suffix: ':SET'
    configuration:
      default_limits: (0, 1)
      values: {'LEVEL0': 1, 'LEVEL1': 0}
    """

    def init(self):
        super().init()
        limits = self._nominal_limits
        EPICSActuator.set_limits(self, limits)
        self._initialise_values()
        current_value = self.get_value()
        self.update_value(current_value)

    def _set_value(self, value):
        if isinstance(value, Enum):
            value = value.value
        super()._set_value(value)

    def get_value(self):
        value = EPICSActuator.get_value(self)
        if not isinstance(value, Enum):
            value = self.value_to_enum(value)
        return value

    def _initialise_values(self):
        values = self.get_property("values")
        self.VALUES = Enum(
            "ValueEnum",
            dict(values, **{item.name: item.value for item in BaseValueEnum}),
        )

    def update_value(self, value=None) -> None:
        if value is None:
            return
        if not isinstance(value, Enum):
            value = self.value_to_enum(value)
        super().update_value(value)

    def hasnt_arrived(self, setpoint):
        if not isinstance(setpoint, Enum):
            setpoint = self.value_to_enum(setpoint)
        readback = self.get_value()
        if not readback:
            return False
        return setpoint != readback


class EPICSNStateInterval(EPICSNState):
    """
    This class is a workaround for devices that exist in a discrete number of states
    and do NOT have an on/off switch, but NEED to have a virtual on/off state in the
    GUI. Example: our frontlight device only allows for intensity setting (between 0
    and 20000), but does not have an on/off switch.

    This class implements the following logic: if the state of the device is higher
    than LEVEL1, it will be shown as on in the GUI. Otherwise, it will be shown as
    off in the GUI.

    Clicking on the icon when it is off will bring it to the LEVEL0 intensity value.
    Clicking on the icon when it is on will bring it to the LEVEL1 intensity value.

    YAML Example
    ------------

    %YAML 1.2
    ---
    class: LNLS.EPICS.EPICSNState.EPICSNStateInterval
    epics:
    "MNC:B:LUCIOLE01:LIGHT_CH2":
      channels:
        rbv:
          suffix: ''
          polling_period: 200
        val:
          suffix: ''
    configuration:
      default_limits: (0, 1)
      values: {'LEVEL0': 15000, 'LEVEL1': 0}
    """

    def update_value(self, value=None) -> None:
        if value is None:
            return
        if isinstance(value, Enum):
            value = value.value
        if not isinstance(value, int):
            return
        values_list = list(self.VALUES)
        if not (len(values_list) > 1):
            return
        enum_value = values_list[0] if value > values_list[1].value else values_list[1]
        super().update_value(enum_value)
